unit ScaleManager;
//   Use this unit to prepare a Scale (a ChartAxis descendant) parallel to an
//  axis on an existing Chart.  The reference axis must have fixed (not
//  Automatic) Minimum & Maximum, and its MarginUnits must be muPercent.

//    This unit also provides the SetIncrement procedure to set the Increment of
//  any axis.

//    Before creating a Scale, Chart must contain a new LineSeries
//  (say, seriesScale).  None of the parameters of seriesScale needs to be
//  edited.

  interface
    uses
      Chart    (* tCustomChart *),
      Controls (* tDate *),
      Graphics (* clBlack, clBlue, clRed, tColor *),
      Series   (* tLineSeries *),
      TeEngine (* tChartAxis *);

    type
      tScaleSide = (sideBottom, sideLeft, sideRight, sideTop);

    type
      tScale = class(tChartAxis)
        private
          ShowingScaleGrid: boolean;
          TheParentAxis: tChartAxis;
          InnerPercent, OuterPercent, TheSubtrahend, TheDivider: real;
          TheParentScaleName, TheScaleName: string;
          procedure SetTitleAndColors(Axis: tChartAxis;
                                      const TheTitle: string;
                                      const Color: tColor);
        public
          procedure   AdjustLimits;
          constructor Create(Chart: tCustomChart;
                             const Side: tScaleSide;
                             const Color: tColor;
                             const Subtrahend, Divider: real;
                             Series: tLineSeries;
                             const ScaleTitle: string;
                             ParentScaleName: string = '';  // for menu
                             ScaleName:       string = ''); // ..
          function    ToggleGrid: string;
        end (* tScale *);

    type
      tBottomDayOffsetScale = class(tScale)
        public
          constructor Create(Chart: tCustomChart; Series: tLineSeries;
                             const Anchor: tDate);
        end (* tBottomDayOffsetScale *);

      tLeftCelsiusScale = class(tScale)
        public
          constructor Create(Chart: tCustomChart; Series: tLineSeries);
        end (* tLeftCelsiusScale *);

    procedure SetIncrement(Axis: tChartAxis);

  implementation
    uses
      Dialogs  (* ShowMessage *),
      SysUtils (* Abort *),
      TeeProcs (* muPercent *);

    procedure SetIncrement(Axis: tChartAxis);
      (* Automatic increment gives increments I don't like *)

      type
        tGrainType = (gt1, gt2, gt5);

      var
        GrainType: tGrainType;
        Grain, ObservedRange: real;

      begin (* SetIncrement *)
        with Axis do
          begin (* Axis *)
            Increment := 10; // avoid degenerate case

            ObservedRange := Maximum - Minimum;
            if ObservedRange > 0 then
              begin (* nondegenerate *)
                GrainType := gt1;
                Grain := 0.0001;
                while Grain < ObservedRange / 5 do
                  begin (* OK, but try for bigger one *)
                    Increment := Grain;
                    case GrainType of
                      gt1: begin (* 0.001, 0.01, 0.1, etc. *)
                             Grain := 2 * Grain;
                             GrainType := gt2
                           end   (* 0.001, 0.01, 0.1, etc. *);
                      gt2: begin (* 0.002, 0.02, 0.2, etc. *)
                             Grain := 2.5 * Grain;
                             GrainType := gt5
                           end   (* 0.002, 0.02, 0.2, etc. *);
                      gt5: begin (* 0.005, 0.05, 0.5, etc. *)
                             Grain := 2 * Grain;
                             GrainType := gt1
                           end   (* 0.005, 0.05, 0.5, etc. *)
                      end (* case *)
                  end   (* OK, but try for bigger one *)
              end   (* nondegenerate *)
          end   (* Axis *)
      end   (* SetIncrement *);

{ tScale }

    procedure tScale.AdjustLimits;
      (* force the scale to sync with the parent axis *)

      begin (* tScale.AdjustLimits *)
        Minimum := -9999;    // guarantee Minimum < Maximum
        Maximum := 99999;    // ..
        Minimum := (TheParentAxis.Minimum - TheSubtrahend) / TheDivider;
        Maximum := (TheParentAxis.Maximum - TheSubtrahend) / TheDivider;
        SetIncrement(self)
      end   (* tScale.AdjustLimits *);

    constructor tScale.Create(Chart: tCustomChart;
                              const Side: tScaleSide;
                              const Color: tColor;
                              const Subtrahend, Divider: real;
                              Series: tLineSeries;
                              const ScaleTitle: string;
                              ParentScaleName, ScaleName: string);
      (* constructor *)
    
      begin (* tScale.Create *)
        Chart.MarginUnits := muPercent;
        case Side of
            sideLeft: begin (* easy case *)
                        Chart.MarginLeft :=  15;
                        InnerPercent     :=  -1;
                        OuterPercent     := -13;
                        TheParentAxis    := Chart.LeftAxis
                      end   (* easy case *);

          sideBottom: begin (* bottom *)
                        Chart.MarginBottom :=  15;
                        InnerPercent       :=  -1;
                        OuterPercent       := -12;
                        TheParentAxis      := Chart.BottomAxis
                      end   (* bottom *);

           sideRight,
             sideTop: begin (* to do *)
                        ShowMessage('Right- & top-side scales not yet implemented');
                        Abort
                      end   (* to do *)
          end (* case *);

        { ancestor properties }
            inherited Create(Chart);
            Automatic        := false;
            AutomaticMaximum := false;
            AutomaticMinimum := false;
            Grid.Visible     := false;
            Horizontal       := TheParentAxis.Horizontal;
            PositionUnits    := muPercent;
            PositionPercent  := OuterPercent;
            Visible          := true;

            SetTitleAndColors(self, ScaleTitle, Color);

        { scale-specific }
            ShowingScaleGrid   := false;
            TheDivider         := Divider;
            TheParentScaleName := ParentScaleName;
            TheScaleName       := ScaleName;
            TheSubtrahend      := Subtrahend;
            AdjustLimits;

        { series-specific }
            with Series do
              begin (* Series *)
                if Horizontal then
                  begin (* special spacing *)
                    CustomHorizAxis := self;
                    LabelsSize       := 2 * LabelsSize
                  end   (* special spacing *)
                else
                  CustomVertAxis  := self;
                  
                SeriesColor    := ParentChart.Color;
                Transparency   := 100;
                ShowInLegend   := false
              end   (* Series *)
      end   (* tScale.Create *);

    procedure tScale.SetTitleAndColors(Axis: tChartAxis;
                                       const TheTitle: string;
                                       const Color: tColor);
      (* utility *)
    
      begin (* tScale.SetTitleAndColors *)
        with Axis do
          begin (* Axis *)
            Axis.Color       := Color;
            Grid.Color       := Color;
            LabelsFont.Color := Color;
            LabelsFont.Size  := 12;
            Title.Font.Color := Color;
            Title.Angle      := 0;
            Title.Font.Size  := 14;
            Title.Font.Style := [fsBold];
            Title.Text       := TheTitle;

            if Horizontal then
              begin (* horizontal *)
                LabelsFont.Size  := 10;
                Title.Font.Size  := 10;
                Title.Font.Style := []
              end   (* horizontal *)
              
            else if length(TheTitle) > 3 then
              Title.Angle := 90
          end   (* Axis *)
      end   (* tScale.SetTitleAndColors *);

    function tScale.ToggleGrid: string;
      (* switch between parent grid and scale grid.  Result is new menu caption *)

      function ShowGrid(Visible, Invisible: tChartAxis;
                        const ToggleMenu: string): string;
        (* move scales, swap grids *)

        begin (* ShowGrid *)
          Visible.Grid.Visible := true;
          Invisible.Grid.Visible := false;
          Visible.PositionPercent := InnerPercent;
          Invisible.PositionPercent := OuterPercent;
          Result := ToggleMenu
        end   (* ShowGrid *);

      begin (* tScale.ToggleGrid *)
        ShowingScaleGrid := not ShowingScaleGrid;
        if ShowingScaleGrid then
          Result := ShowGrid(self, TheParentAxis, TheParentScaleName)
        else
          Result := ShowGrid(TheParentAxis, self, TheScaleName)
      end   (* tScale.ToggleGrid *);

{ tBottomDayOffsetScale }

    constructor tBottomDayOffsetScale.Create(Chart: tCustomChart;
                                             Series: tLineSeries;
                                             const Anchor: tDate);
      (* This method ASSUMES the Chart.BottomAxis shows dates.  This scale
      shows signed offsets (in days) from the Anchor date *)

      const
        clDayOffset = clBlack;

      begin (* tBottomDayOffsetScale.Create *)
        inherited Create(Chart, sideBottom, clDayOffset,
                         Anchor, 1, Series,
                         FormatDateTime('"days from "yyyy-mm-dd', Anchor));
        self.AxisValuesFormat := '"+"###;"-"###;|'
      end   (* tBottomDayOffsetScale.Create *);

{ tLeftCelsiusScale }

   constructor tLeftCelsiusScale.Create(Chart: tCustomChart; Series: tLineSeries);
     (* This method ASSUMES that Chart.LeftAxis is a Fahrenheit scale.  This
     scale shows degrees Celsius *)

     const
       clCelsius  = clBlue;
       clFahrenheit = clRed;

     begin (* tLeftCelsiusScale.Create *)
       inherited Create(Chart, sideLeft, clCelsius,
                        32.0, 1.8, Series,
                        'C',
                        'gridF', 'gridC');

       SetTitleAndColors(Chart.LeftAxis, 'F', clFahrenheit);
     end   (* tLeftCelsiusScale.Create *);

end.
